import React, { Component } from 'react';
import PropTypes from 'prop-types';
import { connect } from 'react-redux';
import { compose } from 'redux';
import { DragDropContext } from 'react-dnd';
import HTML5Backend from 'react-dnd-html5-backend';
import Button from '@material-ui/core/Button';
import Paper from '@material-ui/core/Paper';
import Tabs from '@material-ui/core/Tabs';
import Tab from '@material-ui/core/Tab';
import TextField from '@material-ui/core/TextField';
import Typography from '@material-ui/core/Typography';
import { withStyles } from '@material-ui/core/styles';
import classNames from 'classnames';
import TimeStamp from 'react-timestamp';
import { shell } from 'electron';
import path from 'path';
import Box from './displays/box-plot';
import Scatter from './displays/scatter-plot';
import Table from './displays/result-table';
import Images from './displays/images';
import String from './displays/string';
import PipelineStep from '../pipelines/pipeline-step';
import Iframe from './displays/iframe';
import { getLocalRun } from '../../state/ducks/runs';


const styles = theme => ({
  paper: {
    ...theme.mixins.gutters(),
    paddingTop: theme.spacing.unit * 2,
    paddingBottom: theme.spacing.unit * 2,
    marginTop: theme.spacing.unit * 2,
    display: 'flex',
  },
  resultsInfo: {
    display: 'flex',
    'flex-direction': 'column',
  },
  resultButton: {
    display: 'flex',
    'flex-direction': 'column-reverse',
  },
  timestamp: {
    display: 'flex',
  },
  label: {
    fontWeight: 'bold',
    marginRight: theme.spacing.unit,
  },
  formControl: {
    marginBottom: theme.spacing.unit * 2,
  },
  error: {
    color: 'red',
  },
  errorSmall: {
    color: 'red',
    fontSize: '0.8rem',
  },
});

class Result extends Component {
  constructor(props) {
    super(props);

    this.state = {
      run: {},
      computationOutput: {},
      displayTypes: [],
      type: 'object',
      plotData: [],
      selectedTabIndex: 0,
    };

    this.handleSelect = this.handleSelect.bind(this);
  }

  componentDidMount() {
    const run = this.props.getLocalRun(this.props.params.resultId);

    let plotData = {};

    // Checking display type of computation
    const stepsLength = run.pipelineSnapshot.steps.length;

    let displayTypes = run.pipelineSnapshot.steps[stepsLength - 1]
      .computations[0].computation.display || { type: 'pipeline' };

    this.setState({
      computationOutput: run.pipelineSnapshot.steps[stepsLength - 1]
        .computations[0].computation.output,
      displayTypes,
    });

    if (displayTypes && !displayTypes.length) {
      const array = [];
      array[0] = displayTypes;
      displayTypes = array;
    }

    if (displayTypes && displayTypes.findIndex(disp => disp.type === 'scatter_plot') > -1) {
      plotData.testData = [];
      run.results.plots.map(result => (
        result.coordinates.map(val => (
          plotData.testData.push({
            name: result.title,
            x: val.x,
            y: val.y,
          })
        ))
      ));
    } else if (displayTypes && displayTypes.findIndex(disp => disp.type === 'box_plot') > -1) {
      plotData.testData = [];
      run.results.x.map(val => (
        plotData.testData.push(val)
      ));
    } else {
      plotData = run.results;
    }

    this.setState({
      run,
      plotData,
    });
  }

  handleOpenResult = () => {
    const { run: { id } } = this.state;
    const { auth: { user, appDirectory } } = this.props;
    const resultDir = path.join(appDirectory, 'output', user.id, id);

    shell.openItem(resultDir);
  }

  handleSelect(event, value) {
    this.setState({ selectedTabIndex: value });
  }

  render() {
    const {
      run, selectedTabIndex, plotData, computationOutput,
    } = this.state;
    const { consortia, classes, auth: { appDirectory, user } } = this.props;
    const consortium = consortia.find(c => c.id === run.consortiumId);
    let { displayTypes } = this.state;
    let stepsLength = -1;
    let covariates = [];
    if (run && run.pipelineSnapshot) {
      stepsLength = run.pipelineSnapshot.steps.length;
    }

    if (stepsLength > 0 && run.pipelineSnapshot.steps[stepsLength - 1].inputMap
      && run.pipelineSnapshot.steps[stepsLength - 1].inputMap.covariates) {
      covariates = run.pipelineSnapshot.steps[stepsLength - 1]
        .inputMap.covariates.ownerMappings.map(m => m.name);
    }

    if (displayTypes && !displayTypes.length) {
      const array = [];
      array[0] = displayTypes;
      displayTypes = array;
    }

    const selectedDisplayType = run && run.results
      && displayTypes && displayTypes[selectedTabIndex];

    return (
      <div>
        <Paper className={classes.paper}>
          <div className={classes.resultsInfo}>
            {
              consortium && run.pipelineSnapshot
              && (
                <Typography variant="h6">
                  {`Results: ${consortium.name} || ${run.pipelineSnapshot.name}`}
                </Typography>
              )
            }
            {
              run.startDate
              && (
                <div className={classes.timestamp}>
                  <Typography className={classes.label}>Start date:</Typography>
                  <Typography>
                    <TimeStamp
                      time={run.startDate / 1000}
                      precision={2}
                      autoUpdate={10}
                      format="full"
                    />
                  </Typography>
                </div>
              )
            }
            {
              run.endDate
              && (
                <div className={classes.timestamp}>
                  <Typography className={classes.label}>End date:</Typography>
                  <Typography>
                    <TimeStamp
                      time={run.endDate / 1000}
                      precision={2}
                      autoUpdate={10}
                      format="full"
                    />
                  </Typography>
                </div>
              )
            }
            {
              stepsLength > -1 && covariates.length > 0 && (
                <div>
                  <span className="bold">Covariates: </span>
                  {covariates.join(', ')}
                </div>
              )
            }
          </div>
          <div className={classes.resultButton}>
            <Button
              variant="contained"
              color="primary"
              style={{ marginLeft: 10 }}
              onClick={this.handleOpenResult}
            >
              Open Local Results
            </Button>
          </div>

        </Paper>

        <Tabs
          value={selectedTabIndex}
          onChange={this.handleSelect}
        >
          {
            run && run.results && displayTypes && displayTypes.map((disp) => {
              const title = disp.type.replace('_', ' ')
                .replace(/\w\S*/g, txt => txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase());

              return <Tab key={disp.type} label={`${title} View`} />;
            })
          }
        </Tabs>
        {
          selectedDisplayType
          && (
            <div>
              {
                selectedDisplayType.type === 'string'
                && (
                  <String
                    plotData={plotData}
                    title={`${consortium.name}_${run.pipelineSnapshot.name}`}
                  />
                )
              }
              {
                selectedDisplayType.type === 'box_plot'
                && <Box plotData={plotData.testData} />
              }
              {
                selectedDisplayType.type === 'scatter_plot'
                && <Scatter plotData={plotData.testData} />
              }
              {
                selectedDisplayType.type === 'table'
                && (
                  <Table
                    computationOutput={computationOutput}
                    plotData={plotData}
                    tables={selectedDisplayType.tables ? selectedDisplayType.tables : null}
                    title={`${consortium.name}_${run.pipelineSnapshot.name}`}
                  />
                )
              }
              {
                selectedDisplayType.type === 'iframe'
                && (
                  <Iframe
                    plotData={plotData}
                    title={`${consortium.name}_${run.pipelineSnapshot.name}`}
                    path={`${appDirectory}/output/${user.id}/${run.id}/${run.pipelineSnapshot.steps[0].inputMap.results_html_path.value}`}
                  />
                )
              }
              {
                selectedDisplayType.type === 'images'
                && (
                  <Images
                    plotData={plotData}
                    title={`${consortium.name}_${run.pipelineSnapshot.name}`}
                  />
                )
              }
              {
                selectedDisplayType.type === 'pipeline' && run.pipelineSnapshot
                && (
                  <div>
                    <TextField
                      fullWidth
                      disabled
                      value={run.pipelineSnapshot.name || ''}
                      className={classes.formControl}
                      label="Name"
                    />
                    <TextField
                      fullWidth
                      disabled
                      value={run.pipelineSnapshot.description || ''}
                      className={classes.formControl}
                      label="Description"
                    />
                    {
                      run.pipelineSnapshot.steps.length > 0
                      && run.pipelineSnapshot.steps.map((step, index) => (
                        <PipelineStep
                          computationId={step.computations[0].id}
                          deleteStep={() => {}}
                          eventKey={step.id}
                          id={step.id}
                          key={step.id}
                          moveStep={() => {}}
                          owner={false}
                          pipelineIndex={index}
                          previousComputationIds={
                            run.pipelineSnapshot.steps
                              .filter((s, i) => i < index)
                              .map(s => s.computations[0].id)
                          }
                          step={step}
                          updateStep={() => {}}
                        />
                      ))
                    }
                  </div>
                )
              }
            </div>
          )
        }

        {
          !selectedDisplayType
          || selectedDisplayType.type === ''
          && (
            <Paper className={classNames(classes.paper)}>
              <table>
                <tbody>
                  <tr>
                    <td>
                      <strong>Results Object:</strong>
                    </td>
                  </tr>
                  <tr>
                    <td>
                      {JSON.stringify(run.results)}
                    </td>
                  </tr>
                  <tr>
                    <td>
                      <span className={classNames(classes.errorSmall)}>
                        *Output Type not defined in Compspec.
                      </span>
                    </td>
                  </tr>
                </tbody>
              </table>
            </Paper>
          )
        }

        {
          run && run.error
          && (
            <Paper className={classNames(classes.paper, classes.error)}>
              {JSON.stringify(run.error.message, null, 2)}
            </Paper>
          )
        }
      </div>
    );
  }
}

Result.propTypes = {
  consortia: PropTypes.array.isRequired,
  getLocalRun: PropTypes.func.isRequired,
  params: PropTypes.object.isRequired,
  auth: PropTypes.object.isRequired,
  classes: PropTypes.object.isRequired,
};

const mapStateToProps = ({ auth }) => {
  return { auth };
};

const connectedComponent = compose(
  connect(mapStateToProps, { getLocalRun }),
  DragDropContext(HTML5Backend)
)(Result);

export default withStyles(styles)(connectedComponent);
